package denim.analysis;

import beast.app.treeannotator.TreeSetParser;
import beast.app.util.Arguments;
import beast.core.util.Log;
import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import denim.util.BitUnion;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/*
 * Created by GRJ on 06/06/2017.
 */
public class MigrationAnalyser {



    class HalfMigration {
        int locus;
        BitUnion clade;
        double height;
        boolean matched;

        HalfMigration(BitUnion clade, int locus, double height) {
            this.locus = locus;
            this.clade = clade;
            this.height = height;
            matched = false;
        }
    }


    class Migration {
        int treeIndex;
        int locus;
        BitUnion outClade;
        BitUnion inClade;
        double height;

        Migration(int treeIndex, int locus, BitUnion outClade,  BitUnion inClade, double height) {
            this.treeIndex = treeIndex;
            this.locus = locus;
            this.outClade = outClade;
            this.inClade = inClade;
            this.height = height;
        }
    }


    class GroupedMigration {
        int locus;
        BitUnion outClade;
        BitUnion inClade;
        int count;
        double meanHeight;

        GroupedMigration(int locus, BitUnion outClade,  BitUnion inClade, int count, double meanHeight) {
            this.locus = locus;
            this.outClade = outClade;
            this.inClade = inClade;
            this.count = count;
            this.meanHeight = meanHeight;
        }
    }

    enum OutputType {
        RAW, CLADES
    }


    private static PrintStream progressStream = Log.err;

    private int burninPercentage;
    private String inputFileName;
    private String outputFileName;

    private OutputType outputType;

    List<String> tipLabels;

    String [] alignmentIDs;

    private ArrayList<HalfMigration> outMigs;
    private ArrayList<HalfMigration> inMigs;

    private ArrayList<Migration> migrations;
    private ArrayList<GroupedMigration> groupedMigrations;


    private Tree[] trees;
    private int nTips;


    private MigrationAnalyser(int burninPercentage, OutputType outputType, String inputFileName, String outputFileName ) {
        this.burninPercentage = burninPercentage;
        this.outputType = outputType;
        this.inputFileName = inputFileName;
        this.outputFileName = outputFileName;
    }


    private void fastTreeSet() throws IOException {
        progressStream.println("0              25             50             75            100");
        progressStream.println("|--------------|--------------|--------------|--------------|");
        MyTreeSetParser parser = new MyTreeSetParser(burninPercentage, false);
        Node[] roots = parser.parseFile(inputFileName);
        tipLabels =  parser.getTipLabels();
        trees = new Tree[roots.length];
        int i = 0;
        for (Node root : roots) {
            trees[i++] = new Tree(root);
        }
        alignmentIDs = parser.getAlignmentIDs();
    }



    private void analyseTrees() {
        if (trees.length == 0) {
            progressStream.print("No trees\n");
            return;
        }
        outMigs = new ArrayList<>();
        inMigs = new ArrayList<>();
        migrations = new ArrayList<>();
        groupedMigrations = new ArrayList<>();
        nTips = trees[0].getLeafNodeCount();
        for (int i = 0; i < trees.length; i++) {
            outMigs.clear();
            inMigs.clear();
            Tree tree = trees[i];
            BitUnion [] clades = new BitUnion[tree.getNodeCount()];
            collectHalfMigsInSubtree(tree.getRoot(), clades);
            joinHalfMigrations(i);
        }
        switch (outputType) {
            case RAW:
                writeRawMigrations();
                break;
            case CLADES:
                mergeMigrations();
                writeGroupedMigrations();
                break;
        }
    }


    private void mergeMigrations() {
        if (migrations.size() == 0) {
            return;
        }
        migrations.sort(OUTCLADE_INCLADE_LOCUS_ORDER);
        Migration last = migrations.get(0);
        int m = 1;
        while (m < migrations.size()) {
            int count = 1;
            double meanHeight = last.height;
            while (m < migrations.size()  &&  compareMigrations(last, migrations.get(m)) == 0) {
                count++;
                meanHeight += migrations.get(m).height;
                m++;
            }
            meanHeight /= count;
            groupedMigrations.add(new GroupedMigration(last.locus, last.outClade, last.inClade, count, meanHeight));
            if (m < migrations.size()) {
                last = migrations.get(m);
            }
        }
        groupedMigrations.sort(COUNT_OUTCLADE_INCLADE_LOCUS);
    }



    private void writeRawMigrations() {
        try {
            PrintStream out = new PrintStream(outputFileName);
            out.print("treeIndex locus outClade inClade height\n");
            for (Migration mig : migrations) {
                out.print(mig.treeIndex + " " + mig.locus + " " + mig.outClade.asText() + " " + mig.inClade.asText() + " " + mig.height + "\n");
            }
            out.close();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }


    private void writeGroupedMigrations() {
        try {
            PrintStream out = new PrintStream(outputFileName);
            out.print("tipLabels " + tipLabels.size()+ "\n");
            for (int i = 0; i < tipLabels.size(); i++) {
                out.print(i + " " + tipLabels.get(i) + "\n");
            }
            out.print("alignmentIDs " + alignmentIDs.length + "\n");
            for (int i = 0; i < alignmentIDs.length; i++) {
                out.print(i + " " + alignmentIDs[i] + "\n");
            }
            out.print("count locus outClade inClade meanHeight\n");
            for (GroupedMigration mig : groupedMigrations) {
                out.print(mig.count + " " + mig.locus + " " + mig.outClade.asText() + " " + mig.inClade.asText() + " " + mig.meanHeight + "\n");
            }
            out.close();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }



    private void collectHalfMigsInSubtree(Node node, BitUnion [] clades) {
        if (!node.isLeaf()) {
            collectHalfMigsInSubtree(node.getChild(0), clades);
            collectHalfMigsInSubtree(node.getChild(1), clades);
        }
        BitUnion xx = new BitUnion(nTips);
        if (node.isLeaf()) {
            xx.insert(node.getNr());
            clades[node.getNr()] = xx;
        } else {
            xx.union(clades[node.getChild(0).getNr()]);
            xx.union(clades[node.getChild(1).getNr()]);
        }
        clades[node.getNr()] = xx;
        Object obj = node.getMetaData("migs");
        if (obj instanceof Double[] ) {
            Double [] migs = (Double[]) obj;

            if (!(migs.length % 3 == 0)) {
                System.out.print("DEBUGGING\n");
            }

            assert migs.length % 3 == 0;
            int nMigs = migs.length / 3;
            for (int m = 0; m <nMigs; m++) {
                int locus = (int)(double)migs[3*m];
                assert locus >= 0;
                int change = (int)(double)migs[3*m + 1];
                assert change == 1  ||  change == -1;
                double height = migs[3*m + 2];
                HalfMigration halfMig = new HalfMigration(xx, locus, height);
                if (change < 0) {
                    outMigs.add(halfMig);
                } else {
                    inMigs.add(halfMig);
                }
            }
        }
    }




    private static boolean bitUnionsEqual(BitUnion x1, BitUnion x2) {
        return x1.isContainedIn(x2)  &&  x2.isContainedIn(x1);
    }


    private static int compareBitUnions(BitUnion x1, BitUnion x2) {
        for (int b = 0; b < x1.size(); b++) {
            BitUnion bBit = new BitUnion(x1.size());
            bBit.insert(b);
            boolean bIn1 = bBit.isContainedIn(x1);
            boolean bIn2 = bBit.isContainedIn(x2);
            if (bIn1 != bIn2) {
                return bIn1 ? 1 : -1;
            }
        }
        return 0;
    }


    private void joinHalfMigrations(int treeIndex) {
        assert outMigs.size() == inMigs.size();
        for (int o = 0; o < outMigs.size(); o++) {
            for (int i = 0; i < inMigs.size(); i++) {
                HalfMigration outHM = outMigs.get(o);
                HalfMigration inHM = inMigs.get(i);
                if (!outHM.matched &&  !inHM.matched  &&
                        outHM.height == inHM.height  &&
                        outHM.locus == inHM.locus  &&
                        !bitUnionsEqual(outHM.clade, inHM.clade) ) {
                    outHM.matched = true;
                    inHM.matched = true;
                    Migration mig = new Migration(treeIndex, outHM.locus, outHM.clade, inHM.clade, outHM.height);
                    migrations.add(mig);
                }
            }
        }
    }


    private static int compareMigrations(Migration m1, Migration m2) {
        if (!bitUnionsEqual(m1.outClade, m2.outClade)) {
            return compareBitUnions(m1.outClade, m2.outClade);
        }
        if (!bitUnionsEqual(m1.inClade, m2.inClade)) {
            return compareBitUnions(m1.inClade, m2.inClade);
        }
        return Integer.compare(m1.locus, m2.locus);
    }




    private static int compareGroupedMigrations(GroupedMigration m1, GroupedMigration m2) {
        if (m1.count != m2.count) {
            return (m1.count < m2.count) ? 1 : -1;
        }
        if (!bitUnionsEqual(m1.outClade, m2.outClade)) {
            return compareBitUnions(m1.outClade, m2.outClade);
        }
        if (!bitUnionsEqual(m1.inClade, m2.inClade)) {
            return compareBitUnions(m1.inClade, m2.inClade);
        }
        return Integer.compare(m1.locus, m2.locus);
    }


    private static final Comparator<GroupedMigration> COUNT_OUTCLADE_INCLADE_LOCUS =
            MigrationAnalyser::compareGroupedMigrations;


    private static final Comparator<Migration> OUTCLADE_INCLADE_LOCUS_ORDER =
            MigrationAnalyser::compareMigrations;



    private static void printUsage(Arguments arguments) {
        arguments.printUsage("MigrationAnalyser", "<input-file-name> <output-file-name>");
        progressStream.println();
        progressStream.println("  Example: MigrationAnalyser -burnin 10 test.trees out.txt");
        progressStream.println();
    }


    public static void main(String[] args) throws IOException {

        String inputFileName = null;
        String outputFileName = null;



        Arguments arguments = new Arguments(
                new Arguments.Option[]{
                        new Arguments.IntegerOption("burnin", 0, 99, "the percentage of states to be considered as 'burn-in'"),
                        new Arguments.StringOption("outputtype", "TYPE","Type of output; raw|clades"),
                        new Arguments.Option("help", "option to print this message"),
                });

        try {
            arguments.parseArguments(args);
        } catch (Arguments.ArgumentException ae) {
            progressStream.println(ae);
            printUsage(arguments);
            System.exit(1);
        }

        if (arguments.hasOption("help")) {
            printUsage(arguments);
            System.exit(0);
        }


        int burnin = -1;
        if (arguments.hasOption("burnin")) {
            burnin = arguments.getIntegerOption("burnin");
        }
        if (burnin >= 100) {
            Log.err.println("burnin percentage is " + burnin + " but should be less than 100.");
            System.exit(1);
        }

        OutputType ot = OutputType.CLADES;
        if (arguments.hasOption("outputtype")) {
            String s = arguments.getStringOption("outputtype");
            switch (s) {
                case "raw" :
                    ot = OutputType.RAW;
                    break;
                case "clades" :
                    ot = OutputType.CLADES;
                    break;
            }
        }

        final String[] args2 = arguments.getLeftoverArguments();

        if (args2.length == 2) {
            inputFileName = args2[0];
            outputFileName = args2[1];
        } else {
            if (args2.length > 2) {
                Log.err.println("Unknown option: " + args2[2]);
                Log.err.println();
                printUsage(arguments);
                System.exit(1);
            } else {
                Log.err.println("File name missing");
                Log.err.println();
                printUsage(arguments);
                System.exit(1);
            }
        }
        switch (args2.length) {
            case 2:

                // fall to
            case 1:

                break;
            default: {


            }
        }


        MigrationAnalyser migan = new MigrationAnalyser(burnin, ot, inputFileName, outputFileName);

        try {
            migan.fastTreeSet();
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            e.printStackTrace();
        }

        migan.analyseTrees();

        System.exit(0);





    }




}
